"""Python API bindings for compilation endpoints in the MATE REST API."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional

from mate_common.models.compilations import (
    CompilationInformation,
    CompilationState,
    CompileOptions,
    TargetKind,
    TargetSpecification,
)
from mate_rest_client import artifacts
from mate_rest_client.common import APIModel, Routes

if TYPE_CHECKING:
    import mate_rest_client
    from mate_rest_client import builds


@dataclass(frozen=False, init=False)
class Compilation(APIModel):
    """Represents a compilation retrieved from MATE."""

    def __init__(self, client: mate_rest_client.Client, info: CompilationInformation) -> None:
        self._client = client
        self._info = info

    def refresh(self) -> None:
        """Refresh the internal state of this compilation."""
        refreshed = self._client.compilations.get(self.id_)
        self._info = refreshed._info

    @property
    def id_(self) -> str:
        """Returns a unique identifier for this compilation."""
        return self._info.compilation_id

    @property
    def state(self) -> CompilationState:
        """Returns the state that this compilation is in."""
        return self._info.state

    @property
    def builds(self) -> Iterator[builds.Build]:
        """Yields each build created from this compilation."""
        for build_id in self._info.build_ids:
            yield self._client.builds.get(build_id)

    @property
    def log_artifact(self) -> Optional[artifacts.Artifact]:
        """Returns the compilation log artifact for this compilation, if it has a log."""
        if self._info.log_artifact:
            return artifacts.Artifact(self._client, self._info.log_artifact)
        else:
            return None

    @property
    def source_artifact(self) -> artifacts.Artifact:
        """Returns the source artifact that this compilation was started with."""
        return artifacts.Artifact(self._client, self._info.source_artifact)

    @property
    def artifacts(self) -> Iterator[artifacts.Artifact]:
        """Yields each artifact associated with this compilation."""
        for artifact_id in self._info.artifact_ids:
            yield self._client.artifacts.get(artifact_id)


class CompilationRoutes(Routes):
    """An adapter for interactions with compilation endpoints."""

    def _create(self, spec: TargetSpecification) -> Compilation:
        """Create a compilation directly from a ``TargetSpecification``."""
        info = self._client.post_as(
            CompilationInformation, "/api/v1/compilations", data=spec.json()
        )
        return Compilation(self._client, info)

    def create_from_challenge(
        self,
        options: CompileOptions = CompileOptions(),
        *,
        challenge_name: Optional[str] = None,
        challenge_id: Optional[str] = None,
        target_id: Optional[str] = None,
    ) -> Compilation:
        """Create a new compilation from a challenge, supplied by a CHESS challenge broker.

        Challenges can be identified through one of three mutually exclusive identifiers:

        - ``challenge_name``: the human-friendly "name" of the challenge, like ``example_1``
        - ``challenge_id``: the broker-assigned unique identifier for the challenge
        - ``target_id``: the broker-assigned unique identifier for a particular challenge target

        When ``challenge_name`` or ``challenge_id`` is used, MATE selects the "root"
        target for the specified challenge. When ``target_id`` is used, MATE selects
        that particular target instead. This can be used to specialize the compilation
        pipeline on a particular version of a brokered challenge, e.g. one that's
        been patched to make compilation simpler within MATE.
        """
        if challenge_name is not None:
            kind = TargetKind.BrokeredChallengeName
            handle = challenge_name
        elif challenge_id is not None:
            kind = TargetKind.BrokeredChallengeID
            handle = challenge_id
        elif target_id is not None:
            kind = TargetKind.BrokeredChallengeTargetID
            handle = target_id
        else:
            raise ValueError(
                "exactly one of challenge_name, challenge_id, or target_id must be supplied"
            )

        spec = TargetSpecification(kind=kind, handle=handle, options=options)
        return self._create(spec)

    def create_from_artifact(
        self, artifact: artifacts.Artifact, options: CompileOptions = CompileOptions()
    ) -> Compilation:
        """Create a new compilation from a source artifact.

        The artifact can be any source artifact that MATE knows how to accept. The server will
        reject compilations that are started with a non-source artifact.
        """
        spec = TargetSpecification(kind=TargetKind.Artifact, handle=artifact.id_, options=options)
        return self._create(spec)

    def maybe_get(self, id_: str) -> Optional[Compilation]:
        """Return an compilation by ID, or ``None`` if the compilation does not exist."""
        info = self._client.get_as_maybe(CompilationInformation, f"/api/v1/compilations/{id_}")
        if info is not None:
            return Compilation(self._client, info)
        return None

    def get(self, id_: str) -> Compilation:
        """Retrieves a particular compilation by its unique identifier."""
        info = self._client.get_as(CompilationInformation, f"/api/v1/compilations/{id_}")
        return Compilation(self._client, info)

    def iter(self, **kwargs: Dict[str, Any]) -> Iterator[Compilation]:
        """Yield each compilation currently available.

        Any additional filters can be supplied via keyword arguments.
        """
        resp = self._client.get("/api/v1/compilations", params=dict(**kwargs, detail=True))
        infos = resp.json()
        for info in infos:
            yield Compilation(self._client, CompilationInformation(**info))

    def __iter__(self) -> Iterator[Compilation]:
        """Yields each compilation currently available, regardless of state."""
        yield from self.iter()



"""Python APIs and models for interacting with MATE's REST API."""
from __future__ import annotations

import enum
import functools
import urllib.parse
from dataclasses import dataclass
from typing import Any, Dict, NewType, Optional, Protocol, Type, TypeVar

import requests
from pydantic import BaseModel

from mate_rest_client import analyses, artifacts, builds, compilations, graphs, manticore, pois

_Model = TypeVar("_Model", bound=BaseModel)


class _CurriedRequest(Protocol):
    """A typing protocol to help MyPy understand partial instance methods."""

    def __call__(self, path: str, **kwargs: Any) -> requests.Response:
        ...


class _CurriedRequestAs(Protocol):
    """A typing protocol to help MyPy understand generic partial instance methods."""

    def __call__(self, model: Type[_Model], path: str, **kwargs: Any) -> _Model:
        ...


class _CurriedRequestAsMaybe(Protocol):
    """A typing protocol to help MyPy understand generic partial instance methods."""

    def __call__(self, model: Type[_Model], path: str, **kwargs: Any) -> Optional[_Model]:
        ...


class _Method(str, enum.Enum):
    """HTTP verbs currently needed by the MATE REST API bindings."""

    GET = "GET"
    POST = "POST"
    DELETE = "DELETE"
    PUT = "PUT"


StatusCode = NewType("StatusCode", int)
Url = NewType("Url", str)


@dataclass(frozen=True)
class RestError(Exception):
    """Raised when a request or other operation internal to the REST client fails."""

    status_code: StatusCode
    reason: str
    url: Url

    def __str__(self) -> str:
        return f"{self.status_code}: {self.reason} for {self.url}"

    @classmethod
    def from_(cls, resp: requests.Response) -> RestError:
        """Create a ``RestError`` from the given ``requests.Response``.

        Assumes that the ``request.Response`` is an error response.
        """
        return cls(StatusCode(resp.status_code), resp.reason, Url(resp.url))


@dataclass(frozen=False, init=False)
class Client:
    """The top-level Python client for MATE's REST API.

    Users can interact with subcomponents of the MATE REST API via attributes
    on any ``Client`` instance:

    - ``client.artifacts`` (``artifacts.ArtifactRoutes``): interactions with artifacts
    - ``client.builds`` (``artifacts.BuildRoutes``): interactions with builds
    - ``client.compilations`` (``artifacts.CompilationRoutes``): interactions with compilations
    - ``client.analyses`` (``artifacts.AnalysisRoutes``): interactions with analyses
    - ``client.pois`` (``artifacts.POIRoutes``): interactions with POIs
    - ``client.graphs`` (``artifacts.GraphRoutes``): interactions with graphs
    - ``client.manticore`` (``artifacts.ManticoreRoutes``): interactions with Manticore tasks

    ``Client`` instances can also make direct requests via ``Client.get()``,
    ``Client.post()``, and ``Client.delete()``. This interface should be considered
    lower-level and **not** preferred.
    """

    def __init__(self, base: str = "http://localhost:8666") -> None:
        self._base = base
        self.artifacts = artifacts.ArtifactRoutes(self)
        self.builds = builds.BuildRoutes(self)
        self.compilations = compilations.CompilationRoutes(self)
        self.analyses = analyses.AnalysisRoutes(self)
        self.pois = pois.POIRoutes(self)
        self.graphs = graphs.GraphRoutes(self)
        self.manticore = manticore.ManticoreRoutes(self)

    def _endpoint(self, path: str) -> Url:
        return Url(urllib.parse.urljoin(self._base, path))

    def _request(self, method: _Method, path: str, **kwargs: Dict[str, Any]) -> requests.Response:
        """Make a request, raising ``RestError`` on any failures.

        All keyword arguments are forwarded directly to ``requests.request()``.
        """
        url = self._endpoint(path)
        resp = requests.request(method.value, url, **kwargs)  # type: ignore[arg-type]
        if not resp.ok:
            raise RestError.from_(resp)
        return resp

    def _request_as(
        self, method: _Method, model: Type[_Model], path: str, **kwargs: Dict[str, Any]
    ) -> _Model:
        """Make a request and return the response as the given model, raising ``RestError` on any
        failures.

        All keyword arguments are forwarded directly to ``requests.request()``.
        """
        resp = self._request(method, path, **kwargs)
        return model.parse_raw(resp.text)

    def _request_as_maybe(
        self, method: _Method, model: Type[_Model], path: str, **kwargs: Dict[str, Any]
    ) -> Optional[_Model]:
        """Make a request and return the response as the given model, or ``None`` if the response is
        a 404.

        Raises ``RestError`` on any non-404 failures.

        All keyword arguments are forwarded directly to ``requests.request()``.
        """
        try:
            resp = self._request(method, path, **kwargs)
            return model.parse_raw(resp.text)
        except RestError as re:
            if re.status_code == 404:
                return None
            else:
                raise re

    get: _CurriedRequest = functools.partialmethod(_request, _Method.GET)  # type: ignore[assignment]
    post: _CurriedRequest = functools.partialmethod(_request, _Method.POST)  # type: ignore[assignment]
    delete: _CurriedRequest = functools.partialmethod(_request, _Method.DELETE)  # type: ignore[assignment]
    put: _CurriedRequest = functools.partialmethod(_request, _Method.PUT)  # type: ignore[assignment]

    get_as: _CurriedRequestAs = functools.partialmethod(_request_as, _Method.GET)  # type: ignore[assignment]
    post_as: _CurriedRequestAs = functools.partialmethod(_request_as, _Method.POST)  # type: ignore[assignment]
    delete_as: _CurriedRequestAs = functools.partialmethod(_request_as, _Method.DELETE)  # type: ignore[assignment]
    put_as: _CurriedRequestAs = functools.partialmethod(_request_as, _Method.PUT)  # type: ignore[assignment]

    get_as_maybe: _CurriedRequestAsMaybe = functools.partialmethod(_request_as_maybe, _Method.GET)  # type: ignore[assignment]
    post_as_maybe: _CurriedRequestAsMaybe = functools.partialmethod(_request_as_maybe, _Method.POST)  # type: ignore[assignment]
    delete_as_maybe: _CurriedRequestAsMaybe = functools.partialmethod(_request_as_maybe, _Method.DELETE)  # type: ignore[assignment]
    put_as_maybe: _CurriedRequestAsMaybe = functools.partialmethod(_request_as_maybe, _Method.PUT)  # type: ignore[assignment]



"""Python API bindings for analyses endpoints in the MATE REST API."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Iterator

from mate_common.models.analyses import AnalysisTaskInfo, RegisteredAnalysisInfo
from mate_rest_client.common import APIModel, Routes

if TYPE_CHECKING:
    import mate_rest_client
    from mate_rest_client import builds


@dataclass(frozen=False, init=False)
class Analysis(APIModel):
    """Represents an analysis in MATE."""

    def __init__(self, client: mate_rest_client.Client, info: RegisteredAnalysisInfo) -> None:
        self._client = client
        self._info = info

    @property
    def id_(self) -> str:
        return self._info.analysis_id

    @property
    def filepath(self) -> str:
        return self._info.filepath

    @property
    def name(self) -> str:
        return self._info.name

    @property
    def background(self) -> str:
        return self._info.background

    def run(self, build: builds.Build) -> AnalysisTaskInfo:
        return self._client.post_as(
            AnalysisTaskInfo, f"/api/v1/analyses/{self._info.analysis_id}/run/{build.id_}"
        )


class AnalysisRoutes(Routes):
    """An adapter for interactions with analysis endpoints."""

    def create(self, filename: Path) -> Analysis:
        with filename.open() as io:
            info = self._client.post_as(
                RegisteredAnalysisInfo,
                "/api/v1/analyses",
                files={"analysis": io},
            )

        return Analysis(self._client, info)

    def run(self, build: builds.Build) -> Iterator[AnalysisTaskInfo]:
        resp = self._client.post(f"/api/v1/analyses/run/{build.id_}")
        infos = resp.json()
        for info in infos:
            yield AnalysisTaskInfo(**info)

    def iter(self) -> Iterator[Analysis]:
        """Yield each `Analysis` currently available."""
        resp = self._client.get("/api/v1/analyses")
        infos = resp.json()
        for info in infos:
            yield Analysis(self._client, RegisteredAnalysisInfo(**info))

    def __iter__(self) -> Iterator[Analysis]:
        """Yields each `Analysis` currently available."""
        yield from self.iter()

    @property
    def tasks(self) -> Iterator[AnalysisTaskInfo]:
        resp = self._client.get("/api/v1/analyses/tasks", params=dict(detail=True))
        infos = resp.json()
        for info in infos:
            yield AnalysisTaskInfo(**info)



"""Python API bindings for Manticore endpoints in the MATE REST API."""

from __future__ import annotations

from typing import Any, Dict, Iterator

from mate_common.models.manticore import MantiserveTaskInformation
from mate_rest_client.common import Routes


class ManticoreRoutes(Routes):
    """An adapter for interactions with Manticore endpoints."""

    def iter(self, **kwargs: Dict[str, Any]) -> Iterator[MantiserveTaskInformation]:
        """Yields each `MantiserveTaskInformation` currently available."""
        resp = self._client.get("/api/v1/manticore/tasks", params=dict(**kwargs, detail=True))
        infos = resp.json()
        for info in infos:
            yield MantiserveTaskInformation(**info)

    def __iter__(self) -> Iterator[MantiserveTaskInformation]:
        """Yields each `MantiserveTaskInformation` currently available."""
        yield from self.iter()



"""Python API bindings for POI endpoints in the MATE REST API."""

from __future__ import annotations

from typing import TYPE_CHECKING, Iterator

from mate_common.models.graphs import FlowfinderGraph, FlowfinderNode, GraphRequest, SliceRequest
from mate_rest_client.common import Routes

if TYPE_CHECKING:
    import mate_rest_client
    from mate_rest_client import builds


class GraphRoutes(Routes):
    """An adapter for interactions with Graph endpoints."""

    def get_graph(self, build: builds.Build, req: GraphRequest) -> FlowfinderGraph:
        return self._client.post_as(FlowfinderGraph, f"/api/v1/graphs/{build.id_}", data=req.json())

    def get_node(self, build: builds.Build, node_id: str) -> FlowfinderGraph:
        return self._client.get_as(FlowfinderGraph, f"/api/v1/graphs/{build.id_}/nodes/{node_id}")

    def get_slice(self, build: builds.Build, req: SliceRequest) -> FlowfinderGraph:
        return self._client.post_as(
            FlowfinderGraph, f"/api/v1/graphs/{build.id_}/slices", data=req.json()
        )

    def get_function_nodes(self, build: builds.Build) -> Iterator[FlowfinderNode]:
        resp = self._client.get(f"/api/v1/graphs/{build.id_}/function-nodes")
        infos = resp.json()
        for info in infos:
            yield FlowfinderNode(**info)



"""Python API bindings for artifact endpoints in the MATE REST API."""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional

from mate_common.models.artifacts import ArtifactInformation, ArtifactKind, ArtifactSpecification
from mate_rest_client.common import APIModel, Routes

if TYPE_CHECKING:
    import mate_rest_client
    from mate_rest_client import builds, compilations


@dataclass(frozen=False, init=False)
class Artifact(APIModel):
    """Represents an artifact retrieved from MATE."""

    def __init__(self, client: mate_rest_client.Client, info: ArtifactInformation) -> None:
        self._client = client
        self._info = info

    def refresh(self) -> None:
        """Refresh the internal state of this artifact."""
        refreshed = self._client.artifacts.get(self.id_)
        self._info = refreshed._info

    @property
    def id_(self) -> str:
        """Return the artifact's unique ID."""
        return self._info.artifact_id

    @property
    def kind(self) -> ArtifactKind:
        """Return the artifact's kind."""
        return self._info.kind

    @property
    def has_object(self) -> bool:
        """Return whether or not this artifact has any data currently associated with it."""
        return self._info.has_object

    @property
    def attributes(self) -> Dict[str, Any]:
        """Return all unstructured attributes associated with this artifact."""
        return self._info.attributes

    @property
    def compilations(self) -> Iterator[compilations.Compilation]:
        """Yield each compilation associated with this artifact.

        Most artifacts are only associated with a single compilation but some, like source
        artifacts, can be associated with multiple.
        """
        for compilation_id in self._info.compilation_ids:
            yield self._client.compilations.get(compilation_id)

    @property
    def builds(self) -> Iterator[builds.Build]:
        """Yield each build associated with this artifact.

        Most artifacts are only associated with a single build but some, like bitcode artifacts, can
        be associated with multiple.
        """
        for build_id in self._info.build_ids:
            yield self._client.builds.get(build_id)

    def contents(self) -> bytes:
        """Return the contents of this artifact, as bytes."""
        resp = self._client.get(f"/api/v1/artifacts/{self.id_}/object")
        return resp.content


class ArtifactRoutes(Routes):
    """An adapter for interactions with artifact endpoints."""

    def create(
        self,
        kind: ArtifactKind,
        *,
        filename: Path,
        attributes: Dict[str, Any] = {},
    ) -> Artifact:
        """Create a new artifact on the server and return it."""
        spec = ArtifactSpecification(kind=kind, attributes=attributes)

        # Create the artifact itself.
        info = self._client.post_as(ArtifactInformation, "/api/v1/artifacts", data=spec.json())

        # Upload the given file and associate it with the artifact.
        with filename.open("rb") as io:
            info = self._client.post_as(
                ArtifactInformation,
                f"/api/v1/artifacts/{info.artifact_id}/object",
                files={"file": io},
            )

        return Artifact(self._client, info)

    def maybe_get(self, id_: str) -> Optional[Artifact]:
        """Return an artifact by ID, or ``None`` if the artifact does not exist."""
        info = self._client.get_as_maybe(ArtifactInformation, f"/api/v1/artifacts/{id_}")
        if info is not None:
            return Artifact(self._client, info)
        return None

    def get(self, id_: str) -> Artifact:
        """Return an artifact by ID, raising on any error."""
        info = self._client.get_as(ArtifactInformation, f"/api/v1/artifacts/{id_}")
        return Artifact(self._client, info)

    def iter(self, **kwargs: Dict[str, Any]) -> Iterator[Artifact]:
        """Iterate over all artifacts currently known to MATE.

        Any additional filters can be supplied via keyword arguments.
        """
        resp = self._client.get("/api/v1/artifacts", params=dict(**kwargs, detail=True))
        infos = resp.json()
        for info in infos:
            yield Artifact(self._client, ArtifactInformation(**info))

    def __iter__(self) -> Iterator[Artifact]:
        """Iterate over all artifacts currently known to MATE."""
        yield from self.iter()



#!/usr/bin/env python3

from setuptools import find_packages, setup

with open("./README.md") as f:
    long_description = f.read()

with open("./requirements.txt") as f:
    requirements = list(f.read().splitlines())

setup(
    name="mate-rest-client",
    version="0.0.1-pre.0",
    license="BSD-3-Clause",
    author="The MATE authors",
    author_email="mate@galois.com",
    description="A client for MATE's REST API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url="TODO",
    project_urls={"Documentation": "https://mate.galois.com"},
    packages=find_packages(),
    platforms="any",
    python_requires=">=3.8",
    install_requires=requirements,
)



"""Common classes and routines for MATE's REST client."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Protocol

from pydantic import BaseModel

if TYPE_CHECKING:
    import mate_rest_client


class Routes:
    """A management wrapper for a group of related REST endpoints."""

    def __init__(self, client: mate_rest_client.Client) -> None:
        self._client = client


class _HasInfo(Protocol):
    """A typing protocol for objects that have a backing Pydantic ``BaseModel``."""

    _info: BaseModel


class APIModel(_HasInfo):
    """Common functionality for all models returned by the REST API wrapper."""

    def dict(self) -> Dict[str, Any]:
        return self._info.dict()

    def json(self, *args: str, **kwargs: Any) -> str:
        pretty = kwargs.pop("pretty", False)
        if pretty:
            kwargs["sort_keys"] = True
            kwargs["indent"] = 4
        return self._info.json(*args, **kwargs)



"""Python API bindings for POI endpoints in the MATE REST API."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional

from mate_common.models.analyses import POIDetail, POIResultComplexity, POIResultInfo
from mate_rest_client.common import Routes

if TYPE_CHECKING:
    import mate_rest_client


@dataclass(frozen=False, init=False)
class POI:
    """Represents a POI in MATE."""

    def __init__(self, client: mate_rest_client.Client, info: POIResultInfo) -> None:
        self._client = client
        self._info = info

    def refresh(self) -> None:
        """Refresh the internal state of this POI."""
        for refreshed in self._client.pois:
            if refreshed.id_ == self.id_:
                self._info = refreshed._info
                break

    @property
    def id_(self) -> str:
        """Returns a unique identifier for this POI."""
        return self._info.poi_result_id

    @property
    def build_id(self) -> str:
        """Returns the identifier of the build associated with this POI."""
        return self._info.build_id

    @property
    def task_id(self) -> str:
        """Returns the identifier of the analysis task associated with this POI."""
        return self._info.analysis_task_id

    @property
    def analysis_id(self) -> str:
        """Returns the identifier of the analysis that produced this POI."""
        return self._info.analysis_id

    @property
    def analysis_name(self) -> str:
        """Returns the name of the analysis that produced this POI."""
        return self._info.analysis_name

    @property
    def raw_data(self) -> Dict[str, Any]:
        """Returns the raw POI result."""
        return self._info.poi

    @property
    def flagged(self) -> bool:
        """Returns whether the POI has been marked as ``flagged``."""
        return self._info.flagged

    @flagged.setter
    def flagged(self, val: bool) -> None:
        """Sets the ``flagged`` marker on the POI."""
        self._client.put(f"/api/v1/pois/{self.id_}", params={"flagged": val})
        self.refresh()

    @property
    def done(self) -> bool:
        """Returns whether the POI has been marked as ``done``"""
        return self._info.done

    @done.setter
    def done(self, val: bool) -> None:
        """Sets the ``done`` marker on the POI."""
        self._client.put(f"/api/v1/pois/{self.id_}", params={"done": val})
        self.refresh()

    @property
    def complexity(self) -> POIResultComplexity:
        """Returns the estimated complexity of the POI."""
        return self._info.complexity

    @property
    def parent_id(self) -> Optional[str]:
        """Returns the identifier of the parent POI if the POI was created from a parent."""
        return self._info.parent_result_id

    @property
    def child_ids(self) -> List[str]:
        """Returns any child POIs that were produced as a result of this POI."""
        return self._info.child_result_ids

    @property
    def detail(self) -> POIDetail:
        """Returns a struct containing extra details about the POI."""
        return self._client.get_as(POIDetail, f"/api/v1/pois/{self.id_}")


class POIRoutes(Routes):
    """An adapter for interactions with POI endpoints."""

    def __iter__(self) -> Iterator[POI]:
        """Yields each `POI` currently available."""
        resp = self._client.get("/api/v1/pois")
        infos = resp.json()
        for info in infos:
            yield POI(self._client, POIResultInfo(**info))



"""Python API bindings for build endpoints in the MATE REST API."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional

import mate_rest_client
from mate_common.models.analyses import AnalysisTaskInfo, POIResultInfo
from mate_common.models.builds import BuildInformation, BuildOptions, BuildState
from mate_rest_client import artifacts
from mate_rest_client.common import APIModel, Routes
from mate_rest_client.pois import POI

if TYPE_CHECKING:
    from mate_rest_client import compilations


@dataclass(frozen=False, init=False)
class Build(APIModel):
    """Represents a build retrieved from MATE."""

    def __init__(self, client: mate_rest_client.Client, info: BuildInformation) -> None:
        self._client = client
        self._info = info

    def refresh(self) -> None:
        """Refresh the internal state of this build."""
        refreshed = self._client.builds.get(self.id_)
        self._info = refreshed._info

    @property
    def id_(self) -> str:
        """Returns a unique identifier for this build."""
        return self._info.build_id

    @property
    def state(self) -> BuildState:
        """Returns the state that this build is in."""
        return self._info.state

    @property
    def compilation(self) -> compilations.Compilation:
        """Returns the compilation that this build was produced from."""
        return compilations.Compilation(self._client, self._info.compilation)

    @property
    def bitcode_artifact(self) -> artifacts.Artifact:
        """Returns the bitcode artifact that this build was started with."""
        return artifacts.Artifact(self._client, self._info.bitcode_artifact)

    @property
    def artifacts(self) -> Iterator[artifacts.Artifact]:
        """Yields each artifact associated with this build."""
        for artifact_id in self._info.artifact_ids:
            yield self._client.artifacts.get(artifact_id)

    @property
    def tasks(self) -> Iterator[AnalysisTaskInfo]:
        """Yields each analysis task associated with this build."""
        try:
            resp = self._client.get(f"/api/v1/analyses/tasks/{self.id_}", params=dict(detail=True))
        except mate_rest_client.RestError as re:
            if re.status_code == 404:
                return
            else:
                raise
        infos = resp.json()
        for info in infos:
            yield AnalysisTaskInfo(**info)

    @property
    def pois(self) -> Iterator[POI]:
        """Yields each POI associated with this build."""
        try:
            resp = self._client.get(f"/api/v1/pois/build/{self.id_}")
        except mate_rest_client.RestError as re:
            if re.status_code == 404:
                return
            else:
                raise
        infos = resp.json()
        for info in infos:
            yield POI(self._client, POIResultInfo(**info))


class BuildRoutes(Routes):
    """An adapter for interactions with build endpoints."""

    def create_from_compilation(
        self,
        compilation: compilations.Compilation,
        options: BuildOptions,
        *,
        target: Optional[str] = None,
        run_all_pois: bool = False,
    ) -> List[Build]:
        """Create one or more builds from a compilation.

        If ``target`` is passed and matches a binary produced by the compilation,
        creates a build only for that binary.

        All created builds are returned in a list.
        """
        if target is None:
            resp = self._client.post(
                f"/api/v1/builds/{compilation.id_}/build",
                params=dict(run_all_pois=run_all_pois),
                data=options.json(),
            )
            build_ids = resp.json()
            return [self.get(id_) for id_ in build_ids]
        else:
            info = self._client.post_as(
                BuildInformation,
                f"/api/v1/builds/{compilation.id_}/build/single",
                params=dict(target=target),
                data=options.json(),
            )
            return [Build(self._client, info)]

    def maybe_get(self, id_: str) -> Optional[Build]:
        """Return an build by ID, or ``None`` if the build does not exist."""
        info = self._client.get_as_maybe(BuildInformation, f"/api/v1/builds/{id_}")
        if info is not None:
            return Build(self._client, info)
        return None

    def get(self, id_: str) -> Build:
        """Retrieves a particular ``Build`` by its unique identifier."""
        info = self._client.get_as(BuildInformation, f"/api/v1/builds/{id_}")
        return Build(self._client, info)

    def iter(self, **kwargs: Dict[str, Any]) -> Iterator[Build]:
        """Yield each ``Build`` currently available.

        Any additional filters can be supplied via keyword arguments.
        """
        resp = self._client.get("/api/v1/builds", params=dict(**kwargs, detail=True))
        infos = resp.json()
        for info in infos:
            yield Build(self._client, BuildInformation(**info))

    def __iter__(self) -> Iterator[Build]:
        """Yields each ``Build`` currently available, regardless of state."""
        yield from self.iter()
